-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  04.07.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity uart_tx is -- UART Transmitter
   generic (
      FPGA_SYS_CLK_HZ : integer := 100000000;
      BAUDRATE        : integer := 115200
   );
   port (
      -- System
      clk_i        : in  std_logic := '0';
      rst_i        : in  std_logic := '0';
      
      -- Uart line
      txd_o        : out std_logic := '0';
      
      -- FIFO
      fifo_data_i  : in  std_logic_vector(7 downto 0) := (others => '0');
      fifo_wr_i    : in  std_logic := '0';
      fifo_clr_i   : in  std_logic := '0';
      fifo_empty_o : out std_logic;
      fifo_full_o  : out std_logic;
      fifo_usedw_o : out std_logic_vector (7 downto 0)
   );
end entity uart_tx;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of uart_tx is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   constant CLK_DIV    : integer := (FPGA_SYS_CLK_HZ/BAUDRATE) - 1;
   constant CLK_DIV_14 : integer := CLK_DIV / 4;
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component fifo
      port (
         clock  : in std_logic ;
         data   : in std_logic_vector (7 downto 0);
         rdreq  : in std_logic ;
         sclr   : in std_logic ;
         wrreq  : in std_logic ;
         empty  : out std_logic ;
         full   : out std_logic ;
         q      : out std_logic_vector (7 downto 0);
         usedw  : out std_logic_vector (7 downto 0)
      );
   end component;
   
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   -- Transmitter states
   type   fsm_tx_state_t is (FSM_STATE_IDLE, FSM_STATE_START, FSM_STATE_DATA, FSM_STATE_STOP, FSM_STATE_END);
   signal fsm_tx_state    : fsm_tx_state_t := FSM_STATE_IDLE;
   
   -- Misc
   signal cnt        : integer range 0 to CLK_DIV;
   signal bit_cnt    : integer range 0 to 15 := 0;
   signal data       : std_logic_vector(7 downto 0);
   signal data_rd    : std_logic := '0';
   signal fifo_empty : std_logic := '0';
   signal txd        : std_logic := '1';
   
begin

   --
   -- FSM
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then

         -- Sample clock
         if (cnt < CLK_DIV) then
            cnt <= cnt + 1;
         else
            cnt <= 0;
         end if;   
         
         -- Default
         data_rd <= '0';
         
         case (fsm_tx_state) is
         
            --
            -- FSM_STATE_IDLE
            --
            when FSM_STATE_IDLE =>
               cnt     <= 0;
               bit_cnt <= 0;
               txd     <= '1';
               
               if (fifo_empty = '0') then
                  cnt          <= 0;
                  fsm_tx_state <= FSM_STATE_START;
               end if;
            -- FSM_STATE_IDLE
            
            
            --
            -- FSM_STATE_START
            --
            when FSM_STATE_START =>
               txd <= '0';
               
               if (cnt = CLK_DIV) then
                  txd          <= data(bit_cnt);
                  fsm_tx_state <= FSM_STATE_DATA;
               end if;
            -- FSM_STATE_START
            
            
            --
            -- FSM_STATE_DATA
            --
            when FSM_STATE_DATA =>
               txd <= data(bit_cnt);
               
               if (bit_cnt < 8) then
                  if (cnt = CLK_DIV) then
                     bit_cnt <= bit_cnt + 1;
                  end if;
               else    
                  fsm_tx_state <= FSM_STATE_STOP;
               end if;                  
            -- FSM_STATE_DATA
            
            
            --
            -- FSM_STATE_STOP
            --
            when FSM_STATE_STOP =>
               txd <= '1';
               
               if (cnt = CLK_DIV) then
                  data_rd      <= '1';
                  fsm_tx_state <= FSM_STATE_END;
               end if;
            -- FSM_STATE_STOP
            

            --
            -- FSM_STATE_END
            --
            when FSM_STATE_END =>
               -- Give the receiver some time to relaxe
               if (cnt = CLK_DIV_14) then
                  fsm_tx_state <= FSM_STATE_IDLE;
               end if;   
            -- FSM_STATE_END


            when others =>
               fsm_tx_state <= FSM_STATE_IDLE;

         end case;
      
      end if;
   end process;


   --
   -- TX FIFO
   --
   inst_tx_fifo : fifo
      port map (
         clock  => clk_i,
         
         data   => fifo_data_i,
         wrreq  => fifo_wr_i,  

         q      => data,
         rdreq  => data_rd,
         sclr   => fifo_clr_i,
         empty  => fifo_empty,
         full   => fifo_full_o,
         usedw  => fifo_usedw_o
      );
             

   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- Uart line
   txd_o      <= txd;
   
   -- FIFO
   fifo_empty_o <= fifo_empty;

end architecture syn;

-- *** EOF ***

