/**************************************************************************
*  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de)
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  21.08.2017  mifi  First Version.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "project.h"
#include "rsu.h"
#include "xmodem.h"
#include "xboot.h"
#include "adler32.h"

#include "system.h"
#include "altera_avalon_pio_regs.h"
#include "epcs_commands.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define IMAGE_BUFFER_SIZE  FLASH_END_ADDR  

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static uint8_t ImageBuffer[IMAGE_BUFFER_SIZE] __attribute__ ((aligned (16)));

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  OutputBootMessage                                                    */
/*                                                                       */
/*  Output boot message.                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputBootMessage (void)
{
   const char ResetScreen[] = { 0x1B, 'c', 0 };
   
   term_printf("%s", ResetScreen);
   OSTimeDly(100);   
   
   term_printf("\r\n");
   term_printf("*********************************\r\n");
   term_printf("  Project: %s\r\n", PROJECT_NAME);
   term_printf("  Board  : %s\r\n", PROJECT_BOARD);
   term_printf("  Version: v%s\r\n", PROJECT_VER_STRING);
   term_printf("  Build  : "__DATE__ " " __TIME__"\r\n");
   term_printf("*********************************\r\n");
   term_printf("\r\n");
   
   DB_PRINTF("--- Debug output enabled ---\r\n");
   DB_PRINTF("\r\n");
   
} /* OutputBootMessage */

/*************************************************************************/
/*  OutputUsageInfo                                                      */
/*                                                                       */
/*  Output usage information.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputUsageInfo (void)
{
   term_printf("*** Update Mode ***\r\n");
   term_printf("h: output this info\r\n");
   term_printf("i: display image info\r\n");
   term_printf("s: start image\r\n");
   term_printf("e: erase image\r\n");
   term_printf("x: start XModem update\r\n");
   term_printf("E: erase image area\r\n");
   term_printf("\r\n");
   
} /* OutputUsageInfo */

/*************************************************************************/
/*  DisplayImageInfo                                                     */
/*                                                                       */
/*  Output some image infos.                                             */
/*                                                                       */
/*  In    : dStartAddress                                                */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void DisplayImageInfo (uint32_t dStartAddress)
{
   uint32_t    dBaseAddress = EPCS_FLASH_CONTROLLER_BASE + EPCS_FLASH_CONTROLLER_REGISTER_OFFSET;
   uint32_t    dCRC; 
   XBOOT_HEADER Header;

   /* Read XBOOT header */
   epcs_read_buffer(dBaseAddress, dStartAddress, (uint8_t*)&Header, sizeof(Header));
   
   /* Check header */
   if ((XBOOT_HEADER_MAGIC_1 == Header.dMagic1) &&
       (XBOOT_HEADER_MAGIC_2 == Header.dMagic2) &&
       (XBOOT_HEADER_SIZEVER == Header.dSizeVersion))
   {
      /* Check CRC of the header */
      dCRC = adler32(ADLER_START_VALUE, (const uint8_t *)&Header, sizeof(XBOOT_HEADER) - XBOOT_SIZE_OF_CRC32);
      if (dCRC == Header.dHeaderCRC32)
      {
         term_printf("StartAddress    : 0x%08X\r\n", Header.dStartAddress);
         term_printf("DataTotalSize   : %ld bytes\r\n", Header.dDataTotalSize);
         term_printf("DataCreationTime: %ld\r\n", Header.dDataCreationTime);
         
         if (strlen((char*)Header.DataName) < XBOOT_DATA_NAME_SIZE)
         {
            term_printf("DataName        : %s\r\n", Header.DataName);
         }
      }
      else
      {
         term_printf("Header CRC error\r\n");
      }
   }
   else
   {
      term_printf("No image available\r\n");
   }   

} /* DisplayImageInfo */

/*************************************************************************/
/*  EraseEPCSImage                                                       */
/*                                                                       */
/*  Erase the image in the EPCS.                                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void EraseEPCSImage (void)
{
   uint32_t    dBaseAddress = EPCS_FLASH_CONTROLLER_BASE + EPCS_FLASH_CONTROLLER_REGISTER_OFFSET;
   XBOOT_HEADER Header;

   if (1 == FlashIsImageAvailable())
   {
      /* An valid header is available */
      epcs_read_buffer(dBaseAddress, FLASH_IMG_START_ADDR, (uint8_t*)&Header, sizeof(Header));
      
      term_printf("Erase EPCS image");
      FlashEraseArea(Header.dStartAddress, Header.dDataTotalSize + sizeof(Header));
      term_printf("done\r\n");
   }
   else
   {
      term_printf("No image available\r\n");
   }
   
} /* EraseEPCSImage */

/*************************************************************************/
/*  CheckingBootloaderImage                                              */
/*                                                                       */
/*  Check if the CRC of the bootloader data is valid.                    */
/*                                                                       */
/*  In    : dDataSize, dDataCRC                                          */
/*  Out   : none                                                         */
/*  Return: TRUE / FALSE                                                 */
/*************************************************************************/
static int CheckingBootloaderImage (uint32_t dDataSize, uint32_t dDataCRC)
{
   int      nRes = FALSE;
   uint32_t dBaseAddress = EPCS_FLASH_CONTROLLER_BASE + EPCS_FLASH_CONTROLLER_REGISTER_OFFSET;
   uint32_t dAddress;
   uint32_t dCRC;
   
   term_printf("Checking bootloader image");

   dAddress = 0;
   
   dCRC = ADLER_START_VALUE;
   while(dDataSize)
   {
      if (dDataSize >= FLASH_SECTOR_SIZE)
      {
         epcs_read_buffer(dBaseAddress, dAddress, (uint8_t*)ImageBuffer, FLASH_SECTOR_SIZE);
         
         dCRC = adler32(dCRC, (const uint8_t *)ImageBuffer, FLASH_SECTOR_SIZE);
         
         dAddress  += FLASH_SECTOR_SIZE;
         dDataSize -= FLASH_SECTOR_SIZE;
      }
      else
      {
         epcs_read_buffer(dBaseAddress, dAddress, (uint8_t*)ImageBuffer, dDataSize);

         dCRC = adler32(dCRC, (const uint8_t *)ImageBuffer, dDataSize);
         
         dAddress  += dDataSize;
         dDataSize -= dDataSize;
      }   
      term_printf(".");
   }
   
   /* Check CRC of the data */
   if (dCRC == dDataCRC)
   {
      nRes = TRUE;
      term_printf("OK\r\n");
   }
   else
   {
      term_printf("ERROR\r\n");
   }

   return(nRes);
} /* CheckingBootloaderImage */

/*************************************************************************/
/*  CheckingCRCofEPCSImage                                               */
/*                                                                       */
/*  Check if the CRC of the EPCS image is valid.                         */
/*                                                                       */
/*  In    : dStartAddress                                                */
/*  Out   : none                                                         */
/*  Return: TRUE / FALSE                                                 */
/*************************************************************************/
static int CheckingCRCofEPCSImage (uint32_t dStartAddress)
{
   int         nRes = FALSE;
   uint32_t    dBaseAddress = EPCS_FLASH_CONTROLLER_BASE + EPCS_FLASH_CONTROLLER_REGISTER_OFFSET;
   uint32_t    dCRC; 
   XBOOT_HEADER Header;
   uint32_t    dDataSize;
   uint32_t    dAddress;
   
   term_printf("Checking EPCS image");

   /* Read XBOOT header */
   epcs_read_buffer(dBaseAddress, dStartAddress, (uint8_t*)&Header, sizeof(Header));
   
   /* Check header */
   if ((XBOOT_HEADER_MAGIC_1 == Header.dMagic1) &&
       (XBOOT_HEADER_MAGIC_2 == Header.dMagic2) &&
       (XBOOT_HEADER_SIZEVER == Header.dSizeVersion))
   {
      /* Check CRC of the header */
      dCRC = adler32(ADLER_START_VALUE, (const uint8_t *)&Header, sizeof(XBOOT_HEADER) - XBOOT_SIZE_OF_CRC32);
      if (dCRC == Header.dHeaderCRC32)
      {
         /* The header is valid, read the data into the ImageBuffer */
         dDataSize = Header.dDataTotalSize;
         dAddress  = dStartAddress + sizeof(XBOOT_HEADER);
         
         dCRC = ADLER_START_VALUE;
         while(dDataSize)
         {
            if (dDataSize >= FLASH_SECTOR_SIZE)
            {
               epcs_read_buffer(dBaseAddress, dAddress, (uint8_t*)ImageBuffer, FLASH_SECTOR_SIZE);
               
               dCRC = adler32(dCRC, (const uint8_t *)ImageBuffer, FLASH_SECTOR_SIZE);
               
               dAddress  += FLASH_SECTOR_SIZE;
               dDataSize -= FLASH_SECTOR_SIZE;
            }
            else
            {
               epcs_read_buffer(dBaseAddress, dAddress, (uint8_t*)ImageBuffer, dDataSize);

               dCRC = adler32(dCRC, (const uint8_t *)ImageBuffer, dDataSize);
               
               dAddress  += dDataSize;
               dDataSize -= dDataSize;
            }   
            term_printf(".");
         }
         
         /* Check CRC of the data */
         if (dCRC == Header.dDataCRC32)
         {
            nRes = TRUE;
            term_printf("OK\r\n");
         }
         else
         {
            term_printf("ERROR\r\n");
         }
      }
   }
   else
   {
      term_printf("...ERROR\r\n");
   }       

   return(nRes);
} /* CheckingCRCofEPCSImage */

/*************************************************************************/
/*  HandleImageBuffer                                                    */
/*                                                                       */
/*  Check if the image buffer is valid.                                  */
/*  If the buffer is valid, flash the content.                           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: TRUE / FALSE                                                 */
/*************************************************************************/
static int HandleImageBuffer (void)
{
   int           nRes = FALSE;
   uint32_t      dSize = 0;
   uint32_t      dCRC; 
   XBOOT_HEADER *pHeader;
   
   term_printf("Checking image header...");
   
   pHeader = (XBOOT_HEADER*)ImageBuffer;
   
   /* Check header */
   if ((XBOOT_HEADER_MAGIC_1 == pHeader->dMagic1) &&
       (XBOOT_HEADER_MAGIC_2 == pHeader->dMagic2) &&
       (XBOOT_HEADER_SIZEVER == pHeader->dSizeVersion))
   {
      /* Check CRC of the header */
      dCRC = adler32(ADLER_START_VALUE, (const uint8_t *)pHeader, sizeof(XBOOT_HEADER) - XBOOT_SIZE_OF_CRC32);
      if (dCRC == pHeader->dHeaderCRC32)
      {
         /* CRC is ok, check some values */
         if ((pHeader->dStartAddress + sizeof(XBOOT_HEADER) + pHeader->dDataTotalSize) < FLASH_END_ADDR)
         {
            /* Valid header */
            term_printf("OK\r\n");
         
            term_printf("Checking image data...");
            dCRC = adler32(ADLER_START_VALUE, (const uint8_t *)&ImageBuffer[sizeof(XBOOT_HEADER)], pHeader->dDataTotalSize);
            if (dCRC == pHeader->dDataCRC32)
            {
               term_printf("OK\r\n");
               dSize = sizeof(XBOOT_HEADER) + pHeader->dDataTotalSize;
            }
            else
            {
               term_printf("ERROR\r\n");
            }
         }
         else
         {
            term_printf("ERROR\r\n");
         }            
      }
      else
      {
         term_printf("ERROR\r\n");
      }
   }
   else
   {
      term_printf("ERROR\r\n");
   }
   
   if (dSize != 0)
   {
      if (0 == pHeader->dStartAddress)
      {
         /* Special handling for bootloader update */
         term_printf("Erase bootloader image");
         FlashEraseArea(0, dSize);
         term_printf("done\r\n");
      
         dSize = dSize - sizeof(XBOOT_HEADER);
         term_printf("Program bootloader image");
         FlashProgramArea(0, &ImageBuffer[sizeof(XBOOT_HEADER)], dSize);
         term_printf("done\r\n");
         
         nRes = CheckingBootloaderImage(pHeader->dDataTotalSize, pHeader->dDataCRC32);
         if (TRUE == nRes)
         {
            term_printf("Bootloader ready to start\r\n");
         }
      }
      else
      {
         term_printf("Erase EPCS image");
         FlashEraseArea(pHeader->dStartAddress, dSize);
         term_printf("done\r\n");
      
         term_printf("Program EPCS image");
         FlashProgramArea(pHeader->dStartAddress, ImageBuffer, dSize);
         term_printf("done\r\n");
      
         nRes = CheckingCRCofEPCSImage(pHeader->dStartAddress);
         if (TRUE == nRes)
         {
            term_printf("Image ready to start\r\n");
         }
      }            
   }

   return(nRes);
} /* HandleImageBuffer */

/*************************************************************************/
/*  GetUpdateRequest                                                     */
/*                                                                       */
/*  Check if KEY[0] was pressed.                                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: TRUE / FALSE                                                 */
/*************************************************************************/
static uint8_t GetUpdateRequest (void)
{
   uint8_t  bRequest = FALSE;
   uint32_t dValue;
   
   dValue = ~IORD_ALTERA_AVALON_PIO_DATA(PIO_KEY_BASE) & 0x01;
   if (0x01 == dValue)
   {
      bRequest = TRUE;
   }
   
   return(bRequest);
} /* GetUpdateRequest */

/*************************************************************************/
/*  ImageStartAllowed                                                    */
/*                                                                       */
/*  It takes to long to check if a valid image is available. Therefore   */
/*  the image will be started. The FPGA will only start a valid image.   */
/*  In case of an error the bootloader will be started again, and the    */
/*  error reason ist set by the RSU unit.                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: TRUE / FALSE                                                 */
/*************************************************************************/
static uint8_t ImageStartAllowed (void)
{
   uint32_t dMode;
   uint32_t dError;
   uint8_t  bAllowed = FALSE;

   dMode  = rsu_get_current_state_mode(REMOTE_UPDATE_CONTROLLER_BASE);
   dError = rsu_get_error(REMOTE_UPDATE_CONTROLLER_BASE);

   DB_PRINTF("RSU Mode : 0x%08X\r\n", dMode);
   DB_PRINTF("RSU Error: 0x%08X\r\n", dError);
   DB_PRINTF("\r\n");
   
   if ((0 == dMode) && (0 == dError))
   {
      bAllowed = TRUE;
   }

   return(bAllowed);
} /* ImageStartAllowed */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
int main (void)
{
   int      nResult;
   uint32_t dActualTime;
   uint8_t  bImageStartAllowed = FALSE;
   uint8_t  bImageAvailable    = FALSE;
   uint8_t  bUpdateRequest     = FALSE;
   uint8_t  bStartXModem       = FALSE;

   /*
    * Initialize system, com and terminal functionality
    */
   SystemInit(); 
   
   /* Output startup messages */
   OutputBootMessage();
   
   /* Check if a manual update request is available */
   bUpdateRequest = GetUpdateRequest();

   /* Check if an image start is allowed */
   bImageStartAllowed = ImageStartAllowed();
   
   /* Check if an image is available */
   bImageAvailable = (1 == FlashIsImageAvailable()) ? TRUE : FALSE;  
   
   /* Check if an image can be started */
   if ((FALSE == bUpdateRequest) && (TRUE == bImageAvailable) && (TRUE == bImageStartAllowed))
   {
      FlashStartImage();
   }

   /* Check for update request */   
   if (TRUE == bUpdateRequest)
   {
      term_printf("Update request, enter update mode\r\n");
   }
   else
   {
      /* Check for image error */
      if (FALSE == bImageStartAllowed)
      {
         bStartXModem = TRUE;
         term_printf("Image error\r\n");
      }
      
      if (FALSE == bImageAvailable)
      {
         bStartXModem = TRUE;
         term_printf("No image available\r\n");
      }
   }
   
   term_printf("\r\n");
   
   
   /*
    * Check if XModem transfer must be started
    */   
   if (TRUE == bStartXModem)
   {
      term_printf("Waiting to receive an image, press RETURN to abort\r\n");
      nResult = XMStart(ImageBuffer, IMAGE_BUFFER_SIZE);
      term_printf("\r\n");
      if (TRUE == nResult)
      {
         term_printf("\r\nTransfer finished\r\n");
         HandleImageBuffer();
      }
      else
      {
         term_printf("\r\nTransfer aborted\r\n");
      }
      term_printf("\r\n");   
   }

   /*
    * Update Mode
    */   
   OutputUsageInfo();
   
   while (1)
   {
      dActualTime = OSTimeGet();
      
      term_Task(dActualTime);
   }

  
  /*
   * This return here make no sense.
   * But to prevent the compiler warning:
   * "return type of 'main' is not 'int'
   * we use an int as return :-)
   */ 
  return(0);
} /* main */

/*************************************************************************/
/*  term_RxCallback                                                      */
/*                                                                       */
/*  Will be called from TermTask in case a char is received.             */
/*                                                                       */
/*  In    : bData                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void term_RxCallback (uint8_t bData)
{
   int nResult;
   
   switch (bData)
   {
      case 'h':
      {
         OutputUsageInfo();
         break;
      } 
      
      /****************************************************/
      
      case 'i':
      {
         DisplayImageInfo(FLASH_IMG_START_ADDR);
         term_printf("\r\n");
         break;
      }

      case 'e':
      {
         EraseEPCSImage();
         term_printf("\r\n");
         break;
      }
      
      case 's':
      {
         if (1 == FlashIsImageAvailable())
         {
            FlashStartImage();
         }
         else
         {
            term_printf("No image available\r\n");
         }
         term_printf("\r\n");
         break;
      }
      
      case 'x':
      {
         term_printf("Waiting to receive an image, press RETURN to abort\r\n");
         nResult = XMStart(ImageBuffer, IMAGE_BUFFER_SIZE);
         term_printf("\r\n");
         if (TRUE == nResult)
         {
            term_printf("\r\nTransfer finished\r\n");
            HandleImageBuffer();
         }
         else
         {
            term_printf("\r\nTransfer aborted\r\n");
         }
         term_printf("\r\n");   
         break;
      }
      
      case 'E':
      {
         term_printf("Flash erase image area\r\n");
         FlashEraseAll();
         term_printf("Flash erase done\r\n");
         term_printf("\r\n");
         break;
      }
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (bData) */
} /* term_RxCallback */

/*** EOF ***/
